%{
===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
CT 5/2012 
PURPOSE: this script tests Yale B lighting Faces Recogntion and Analogy

INPUT:  
OUTPUT:  
NOTES:       
TESTED:
CHANGELOG:
TODO:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%}

%%
clear;
addpath('./MTA');
addpath('./util');
addpath('./tensor_toolbox_2.5');


% load data
rand('seed', 1785);
IMSIZE = 24;
[ YaleData YaleTest] = load_ExtendedYaleB(11:38, 1:10, IMSIZE);

%% display training data
ims =[];
for bb = 11:length(YaleData)
    temp = data_shuffle(YaleData{bb}.image);
    ims = [ims; temp(1:10,:)];
end
im = write_grid_images( ims,  IMSIZE*[1 1], [28 10], 2, 0);
myf(50); imshow(im, [0 1])

if 1
    personid = 11;
    im = write_grid_images( YaleData{personid}.image,  IMSIZE*[1 1], [6 5], 2, 0);
    myf(51); imshow(im, [0 1])
end

%% data normalization
global xxTrain;
global xxTest;
global xxValid;

faces = cat(1, YaleData{[11:end]});
xxTrain = double(cat(1, faces(:).image));
train_labels = double(cat(1, faces(:).y));
light_labels = double(cat(1, faces(:).lighting));

faces = cat(1, YaleTest{1:10});
xxTest = double(cat(1, faces(:).image));

datamean = mean(xxTrain,1);
datastd = 1+0*std(xxTrain);
mi(datastd);

xxTrain = bsxfun(@rdivide, bsxfun(@minus, xxTrain, datamean), datastd);
xxTest = bsxfun(@rdivide, bsxfun(@minus, xxTest, datamean), datastd);

%display data again
temp = data_shuffle( xxTrain);
temp = bsxfun(@plus, bsxfun(@times, temp, datastd), datamean);
im = write_grid_images(temp, [IMSIZE IMSIZE], [10 10], 1, 0);
myfclf(52,  1, 1, 0, 0); imshow(im, []);

%split half to validation
randinds = randperm(size(xxTest,1));
xxValid = xxTest(randinds(1:int32(length(randinds)/2)),:);
xxTest = xxTest( randinds(int32(length(randinds)/2)+1:end),:);

%% MTA training
randn('seed',1234);
rand('seed',1234);

tic;
D=size(xxTrain, 2);
d = [80 4];

global labels;
labels=[];
labels{1} = train_labels'-10;
labels{2} = light_labels';

C = 1;

paramsmta.d = d;

paramsmta.C = C;
paramsmta.iso  = 0;
paramsmta.eq =0;
paramsmta.GPU =0;
paramsmta.epsilon =0.001;
paramsmta.QTHRESH = 1e-5;
paramsmta.MIN_PSI = 1e-8;
paramsmta.ZM = 0;
paramsmta.max_iters = 7;
paramsmta.num_mbs = 1;
paramsmta.wtcost = 0.01;

PsiInitFactor = 10;
RandInitFactor = 0.01;

Cntrs = repmat( mean(xxTrain), [C, 1]);
Cntrs = Cntrs + 0.01*randn(size(Cntrs));

paramsmta.PreWts.Psi = 10+0*repmat( PsiInitFactor*(var(xxTrain, 0, 1)'), 1, C);
paramsmta.PreWts.M = Cntrs';

paramsmta.PreWts.W = RandInitFactor*randn(D, sum(d), C);
paramsmta.PreWts.T = RandInitFactor*randn(D, prod(d), C);

paramsmta.PreWts.mix = ones(C,1)/C;

paramsmta.nGibbsSamples = 20;
paramsmta.nLLSamples = 2000;
paramsmta.nUpdateIteratons = 1;
paramsmta.learnrate= 0.8;
paramsmta.EarlyStop = 0;

[logl_epoch MoTA] = mta3(paramsmta);
fprintf('\n mta3 learning done\n');
mi(MoTA.Psi(:))
toc;



%%  --- MTA3 reconstruction samples ----
ReconSamples = 300;

%x_small = xxTrain([1:10:1250],:);
x_small = xxTest([1:2:224],:);

[D N] = size(x_small');

C = length(MoTA.mix);
zs = cell(C,length(MoTA.d));
for cc = 1:C
    for dd =1:length(MoTA.d)  
        zs{cc,dd} = zeros(MoTA.d(dd), N);        
    end
end

for i = 1:ReconSamples
    fprintf('%d.',i)
    for dd = 1:length(MoTA.d)
        
        setinds = [];
        setinds{1} = [1:dd-1 dd+1:length(MoTA.d)];
        setinds{2} = dd;
       
        [ LogL, zs ] = postinf_mta_given_zs( double(x_small)', MoTA, zs, setinds, 1-(i > ReconSamples-10) );       
    end
end
fprintf('\nDone.')

% display reconstruction
c =1;
bNoise = 0;
[recons] = mta3_cond_zs( MoTA, zs, c, bNoise);

myfclf(203, 1, 1, 0,0); hold on;
xx = bsxfun(@plus, bsxfun(@times, x_small, datastd), datamean);
im = write_grid_images(xx, [IMSIZE IMSIZE], [10,10], 1, 0);
imshow(im, [0 1], 'InitialMagnification', 'fit');

myf(204, 1, 1, 0,0); hold on;
recons = recons';
recons = bsxfun(@plus, bsxfun(@times, recons, datastd), datamean);
im = write_grid_images(recons, [IMSIZE IMSIZE], [10 10], 1, 0);
imshow(im, [0 1], 'InitialMagnification', 'fit');

maxabs(recons, xx)


%% Second, recognition using MTA
randn('seed',1234);
rand('seed', 1234);
%if we can use the latent vectors
train_image_ind = 1;
ReconSamples = 200;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Grab Training/Testing Data
SingleTrainX = [];
SingleTrainY = [];
TestX = [];
TestY = [];

nIMAGES =length(train_image_ind);

for pp = 1:length(YaleTest)
    if isempty(YaleTest{pp})
        continue; %no image available, we skip this image
    end
    
    nLight = size(YaleTest{pp}.image, 1);
    images = YaleTest{pp}.image;
    
    SingleTrainX{pp} = images(train_image_ind,:);
    SingleTrainY{pp} = YaleTest{pp}.y(train_image_ind);
    
    test_image_ind = zeros(1, nLight);
    test_image_ind( 8:nLight ) = 1;
    
    TestX{pp} = images( logical(test_image_ind),:);
    TestY{pp} = YaleTest{pp}.y(logical(test_image_ind));
end

GalleryData = cat(1, SingleTrainX{:});
GalleryData = bsxfun(@rdivide, bsxfun(@minus, GalleryData, datamean), datastd);



[D N] = size(GalleryData');

C = length(MoTA.mix);
zs = cell(C,length(MoTA.d));
for cc = 1:C
    for dd =1:length(MoTA.d)  
        zs{cc,dd} = zeros(MoTA.d(dd), N);        
    end
end

test_gallery_labels= ones(1, N);

for i = 1:ReconSamples
    fprintf('%d.',i)
    if rem( i, 50 ) ==0
        fprintf('\n');
    end
    
    for dd = 1:length(MoTA.d)
        
        setinds = [];
        setinds{1} = [1:dd-1 dd+1:length(MoTA.d)];
        setinds{2} = dd;        
        [ LogL, zs ] = postinf_mta_given_zs( double(GalleryData)', MoTA,...
            zs, setinds, 1-(i > ReconSamples-30) );
    end
end
fprintf('\nDone.')


% display reconstruction
c =1;
bNoise = 0;
[recons] = mta3_cond_zs( MoTA, zs, c, bNoise);

myfclf(203, 1, 1, 0,0); hold on;
xx = bsxfun(@plus, bsxfun(@times, GalleryData, datastd), datamean);
im = write_grid_images(xx, [IMSIZE IMSIZE], [1,10], 1, 0);
imshow(im, [0 1], 'InitialMagnification', 'fit');

myf(204, 1, 1, 0,0); hold on;
recons = recons';
recons = bsxfun(@plus, bsxfun(@times, recons, datastd), datamean);
im2 = write_grid_images(recons, [IMSIZE IMSIZE], [1 10], 1, 0);
imshow(im2, [0 1], 'InitialMagnification', 'fit');

maxabs(recons, xx);

myfclf(1000);
imshow( [im;im2], [0 1]);
%

setinds = cell(1,2);
setinds{1} = 1;
setinds{2} = 2;

% inds into the W required.
this_z_inds = false(1, sum(MoTA.d));
rest_z_inds = false(1, sum(MoTA.d));

ind = 1;
for b = 1:length(MoTA.d)
    if b == setinds{2}
        this_z_inds(ind:ind+MoTA.d(b)-1) = true;      
        rest_z_inds(ind:ind+MoTA.d(b)-1) = false;
    else
        this_z_inds(ind:ind+MoTA.d(b)-1) = false;
        rest_z_inds(ind:ind+MoTA.d(b)-1) = true;
    end
    ind = ind + MoTA.d(b);
end

% we now transfer the factor loadings from the closest face found in the
% training set
c=1;
FA_Transfered = [];
for i=1:length(TestX)
    
    FA_Transfered{i}.mix = 1;
    FA_Transfered{i}.Psi = MoTA.Psi;
        
    %split W:
    W_this_z = MoTA.W(:, this_z_inds, c);    %2-biases
    W_rest_z = MoTA.W(:, rest_z_inds, c);    %to be multiplied with z_given
    
    y = cat(1, zs{c,setinds{1}});
    y = y(:,i);
    
    z_c = cell(length(MoTA.d),1);
    [z_c{:}] = zs{c,:};    
    W_batch = bsxfun(@plus, W_this_z, TensorMatrixProd_zs( MoTA.T(:,:,c), MoTA.d, z_c, setinds{1}, setinds{2} ));    
    FA_Transfered{i}.W = W_batch(:,:,i);
    FA_Transfered{i}.M = bsxfun(@plus, MoTA.M(:,c), W_rest_z*y);
    
    if i == 5
        N =100;
        MFA_samples = sample_mfa(FA_Transfered{i}, 100, 0, 2);
        myfclf(6); hold on;

        MFA_samples = bsxfun(@times, bsxfun(@plus, MFA_samples', datamean), datastd);
        im = write_grid_images(MFA_samples, [IMSIZE IMSIZE], [1 8], 1, 1);
        imshow(im, [0 1]); 

        myimshg(SingleTrainX{i}, 5, [0 1]);
    end
end

%

% now perform face recogntion on Test-set

TestXData = cat(1, TestX{:});
TestYData = cat(1, TestY{:});
TestXData = bsxfun(@rdivide, bsxfun(@minus, TestXData, datamean), datastd);

LogLTest = zeros(size(TestXData,1), length(FA_Transfered));
for nn=1:length(FA_Transfered)
    [ LogLTest(:,nn), proj, lat_covs ] = ll_mfa( double(TestXData'), FA_Transfered{nn} );
end

[val subject] = max(LogLTest,[],2);
recogntion_err = 1-mean(subject == TestYData) 



%%  Produces MCMC Trace Plots

C = length(MoTA.mix);
zs = cell(C,length(MoTA.d));
for cc = 1:C
    for dd =1:length(MoTA.d)  
        zs{cc,dd} = zeros(MoTA.d(dd), 2);        
    end
end

ReconSamples = 100;
SAVED = zeros(ReconSamples,6);

for i = 1:ReconSamples
    fprintf('%d.',i)
    if rem( i, 50 ) ==0
        fprintf('\n');
    end
    
    for dd = 1:length(MoTA.d)
                
        SAVED(i, (dd-1)*3+1:dd*3 ) = zs{dd}(1:3,1)';
        
        setinds = [];
        setinds{1} = [1:dd-1 dd+1:length(MoTA.d)];
        setinds{2} = dd;
            
        [ LogL, zs ] = postinf_mta_given_zs( double(TestXData(4:5,:))', MoTA, zs, setinds, 1 );
        
    end
end
fprintf('\nDone.')

myfclf(1); hold on;
plot(0:99, SAVED(:,1), 'r-', 'LineWidth', 2);
plot(0:99, SAVED(:,2), 'g-', 'LineWidth', 2);
plot(0:99, SAVED(:,3), 'b-', 'LineWidth', 2);
plot(0:99, SAVED(:,4), '-', 'Color', [.8 .7 .1],'LineWidth', 2);
plot(0:99, SAVED(:,5), 'm-', 'LineWidth', 2);
plot(0:99, SAVED(:,6), 'k-', 'LineWidth', 2);
title('MCMC trace plot', 'FontSize', 16)
xlabel('Number of Gibbs iterations', 'FontSize', 14)
ylabel('Latent factor values', 'FontSize', 14)
axis([0 100 -5 5]);
box on;
set(gcf, 'Color', [1 1 1 ]);

